// Copyright (c) 2024 The YAC Authors
//
// SPDX-License-Identifier: BSD-3-Clause

#include <stdio.h>
#include <string.h>
#ifdef YAC_NETCDF_ENABLED
#include <netcdf.h>
#endif

#include "tests.h"
#include "test_common.h"
#include "io_utils.h"

void write_dummy_grid_file(
  char * grid_name, char * grid_filename, char * mask_filename,
  int with_corners, size_t num_lon, size_t num_lat,
  double lon_range[2], double lat_range[2]) {

#ifndef YAC_NETCDF_ENABLED
  UNUSED(grid_name);
  UNUSED(grid_filename);
  UNUSED(mask_filename);
  UNUSED(with_corners);
  UNUSED(num_lon);
  UNUSED(num_lat);
  UNUSED(lon_range);
  UNUSED(lat_range);
  die("ERROR(write_dummy_grid_file): YAC is built without the NetCDF support");
#else

  { // grid file
    int ncid;

    // create file
    yac_nc_create(grid_filename, NC_CLOBBER, &ncid);

    char crn_dim_name[128];
    char x_dim_name[128];
    char y_dim_name[128];

    sprintf(crn_dim_name, "crn_%s", grid_name);
    sprintf(x_dim_name, "x_%s", grid_name);
    sprintf(y_dim_name, "y_%s", grid_name);

    int dim_crn_id = -1;
    int dim_x_id;
    int dim_y_id;

    // define dimensions
    if (with_corners)
      YAC_HANDLE_ERROR(nc_def_dim(ncid, crn_dim_name, 4, &dim_crn_id));
    YAC_HANDLE_ERROR(nc_def_dim(ncid, x_dim_name, num_lon, &dim_x_id));
    YAC_HANDLE_ERROR(nc_def_dim(ncid, y_dim_name, num_lat, &dim_y_id));

    char cla_var_name[128];
    char clo_var_name[128];
    char lat_var_name[128];
    char lon_var_name[128];

    sprintf(cla_var_name, "%s.cla", grid_name);
    sprintf(clo_var_name, "%s.clo", grid_name);
    sprintf(lat_var_name, "%s.lat", grid_name);
    sprintf(lon_var_name, "%s.lon", grid_name);

    int corner_dim_ids[3] = {dim_crn_id, dim_y_id, dim_x_id};
    int cell_dim_ids[2] = {dim_y_id, dim_x_id};

    int var_cla_id = -1;
    int var_clo_id = -1;
    int var_lat_id;
    int var_lon_id;

    char degree[] = "degree";
    char title[] = "This is a reg lon-lat dummy grid";

    // define variable
    if (with_corners) {
      YAC_HANDLE_ERROR(
        nc_def_var(
          ncid, cla_var_name, NC_DOUBLE, 3, corner_dim_ids, &var_cla_id));
      YAC_HANDLE_ERROR(
        nc_put_att_text(ncid, var_cla_id, "units", strlen(degree), degree));
      YAC_HANDLE_ERROR(
        nc_put_att_text(ncid, var_cla_id, "title", strlen(title), title));

      YAC_HANDLE_ERROR(
        nc_def_var(
          ncid, clo_var_name, NC_DOUBLE, 3, corner_dim_ids, &var_clo_id));
      YAC_HANDLE_ERROR(
        nc_put_att_text(ncid, var_clo_id, "units", strlen(degree), degree));
      YAC_HANDLE_ERROR(
        nc_put_att_text(ncid, var_clo_id, "title", strlen(title), title));
    }

    YAC_HANDLE_ERROR(
      nc_def_var(
        ncid, lat_var_name, NC_DOUBLE, 2, cell_dim_ids, &var_lat_id));
    YAC_HANDLE_ERROR(
      nc_put_att_text(ncid, var_lat_id, "units", strlen(degree), degree));
    YAC_HANDLE_ERROR(
      nc_put_att_text(ncid, var_lat_id, "title", strlen(title), title));

    YAC_HANDLE_ERROR(
      nc_def_var(
        ncid, lon_var_name, NC_DOUBLE, 2, cell_dim_ids, &var_lon_id));
    YAC_HANDLE_ERROR(
      nc_put_att_text(ncid, var_lon_id, "units", strlen(degree), degree));
    YAC_HANDLE_ERROR(
      nc_put_att_text(ncid, var_lon_id, "title", strlen(title), title));


    // end definition
    YAC_HANDLE_ERROR(nc_enddef(ncid));

    // write grid data

    double cla[4][num_lat][num_lon];
    double clo[4][num_lat][num_lon];
    double lat[num_lat][num_lon];
    double lon[num_lat][num_lon];

    for (size_t i = 0; i < num_lon; ++i) {
      double vertex_lon[2] =
        {((lon_range[1] - lon_range[0]) * (double)i)/(double)num_lon,
         ((lon_range[1] - lon_range[0]) * (double)(i+1))/(double)num_lon};
      for (size_t j = 0; j < num_lat; ++j) {
        double vertex_lat[2] =
            {((lat_range[1] - lat_range[0]) * (double)j)/(double)num_lat,
            ((lat_range[1] - lat_range[0]) * (double)(j+1))/(double)num_lat};
        cla[0][j][i] = lat_range[0] + vertex_lat[0];
        cla[1][j][i] = lat_range[0] + vertex_lat[0];
        cla[2][j][i] = lat_range[0] + vertex_lat[1];
        cla[3][j][i] = lat_range[0] + vertex_lat[1];
        clo[0][j][i] = lon_range[0] + vertex_lon[0];
        clo[1][j][i] = lon_range[0] + vertex_lon[1];
        clo[2][j][i] = lon_range[0] + vertex_lon[1];
        clo[3][j][i] = lon_range[0] + vertex_lon[0];
        lat[j][i] = lat_range[0] + (vertex_lat[0] + vertex_lat[1]) * 0.5;
        lon[j][i] = lon_range[0] + (vertex_lon[0] + vertex_lon[1]) * 0.5;
      }
    }


    if (with_corners) {
      YAC_HANDLE_ERROR(nc_put_var_double(ncid, var_cla_id, &cla[0][0][0]));
      YAC_HANDLE_ERROR(nc_put_var_double(ncid, var_clo_id, &clo[0][0][0]));
    }
    YAC_HANDLE_ERROR(nc_put_var_double(ncid, var_lat_id, &lat[0][0]));
    YAC_HANDLE_ERROR(nc_put_var_double(ncid, var_lon_id, &lon[0][0]));

    YAC_HANDLE_ERROR(nc_close(ncid));
  }

  { // mask file
    int ncid;

    // create file
    yac_nc_create(mask_filename, NC_CLOBBER, &ncid);

    char x_dim_name[128];
    char y_dim_name[128];

    sprintf(x_dim_name, "x_%s", grid_name);
    sprintf(y_dim_name, "y_%s", grid_name);

    int dim_x_id;
    int dim_y_id;

    // define dimensions
    YAC_HANDLE_ERROR(nc_def_dim(ncid, x_dim_name, num_lon, &dim_x_id));
    YAC_HANDLE_ERROR(nc_def_dim(ncid, y_dim_name, num_lat, &dim_y_id));

    char frc_var_name[128];
    char msk_var_name[128];

    sprintf(frc_var_name, "%s.frc", grid_name);
    sprintf(msk_var_name, "%s.msk", grid_name);

    int dim_ids[2] = {dim_y_id, dim_x_id};

    int var_frc_id;
    int var_msk_id;

    char adim[] = "adim";

    // define variable
    YAC_HANDLE_ERROR(
      nc_def_var(
        ncid, frc_var_name, NC_DOUBLE, 2, dim_ids, &var_frc_id));
    YAC_HANDLE_ERROR(
      nc_put_att_text(ncid, var_frc_id, "units", strlen(adim), adim));

    YAC_HANDLE_ERROR(
      nc_def_var(
        ncid, msk_var_name, NC_INT, 2, dim_ids, &var_msk_id));
    YAC_HANDLE_ERROR(
      nc_put_att_text(ncid, var_msk_id, "units", strlen(adim), adim));


    // end definition
    YAC_HANDLE_ERROR(nc_enddef(ncid));

    // write grid data

    double frc[num_lat][num_lon];
    int msk[num_lat][num_lon];

    for (size_t i = 0; i < num_lon; ++i) {
      for (size_t j = 0; j < num_lat; ++j) {
        frc[j][i] = 1;
        msk[j][i] = 0;
      }
    }

    YAC_HANDLE_ERROR(nc_put_var_double(ncid, var_frc_id, &frc[0][0]));
    YAC_HANDLE_ERROR(nc_put_var_int(ncid, var_msk_id, &msk[0][0]));

    YAC_HANDLE_ERROR(nc_close(ncid));
  }
#endif // YAC_NETCDF_ENABLED
}
